import copy
import os
import uuid
from collections.abc import MutableMapping
from collections import OrderedDict
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from chemotion_api.labImotion.items.options import FieldType


class GenericFildValidator:
    def __init__(self, json_data, mapping):
        from chemotion_api.labImotion.items.options import FieldType
        self.Types = FieldType
        self._json_data = json_data
        self._mapping = mapping

    @staticmethod
    def _extract_value(value):
        if isinstance(value, dict) and 'value' in value:
            return value['value']
        return value


    def validate(self, layer, field, value):
        field_obj = self._get_json_field(layer, field)
        if field_obj is None:
            return False
        ft = self.Types(field_obj['type'])
        if self._extract_value(value) is None:
            return None
        if ft == self.Types.INTEGER:
            return int(self._extract_value(value))
        if ft == self.Types.CHECKBOX:
            return bool(self._extract_value(value))
        if ft == self.Types.TEXT:
            return str(self._extract_value(value))
        if ft == self.Types.SYSTEM_DEFINED:
            if isinstance(value, (float, int)):
                value = {'value': value}
            return UnitDict(**value, unit=field_obj['value_system'])
        if ft == self.Types.SELECT:
            value = self._extract_value(value)
            options = field_obj['option_layers']
            options = self._json_data['properties_release']['select_options'][options]['options']
            if isinstance(value, int):
                return options[value]['label']
            return next((x['label'] for x in options if x['label'] == value or x['key'] == value), None)

        return value

    def _get_json_field(self, layer, field):
        temp_segment = self._mapping[layer]
        return next((x for x in self._json_data['properties']['layers'][temp_segment['__key']]['fields'] if
                     x['label'] == field), None)

    def field_type(self, layer, field) -> 'FieldType':
        field_obj = self._get_json_field(layer, field)
        return self.Types(field_obj['type'])

    def convert_keys_to_labels(self, layer, field):
        layer_label = next((layer_label for layer_label, layer_value in self._mapping.items() if
                            isinstance(layer_value, dict) and layer_value['__key'] == layer), None)
        field = next((x for x in self._json_data['properties']['layers'][layer]['fields'] if x['field'] == field), None)
        return layer_label, field['label']

    def convert_labels_to_keys(self, layer, field):
        label = self._mapping[layer]['__key']
        field = self._mapping[layer][field]
        return label, field

class PropertyDict(MutableMapping):

    def __init__(self, on_change, kwargs):
        self.data = kwargs
        self._on_change = on_change

    @property
    def on_change(self):
        return self._on_change

    @on_change.setter
    def on_change(self, value):
        self._on_change = value

    def __getitem__(self, name):
        ret_val = self.data.__getitem__(name)
        if isinstance(ret_val, dict):
            ret_val = PropertyDict(None, ret_val)
        if isinstance(ret_val, PropertyDict):
            def inner_on_change(inner_dict_obj, inner_name, inner_value):
                return self._on_change([self.to_dict(), *inner_dict_obj], [name, *inner_name],
                                       [ret_val.to_dict(), *inner_value])

            ret_val.on_change = inner_on_change
        return ret_val

    def __setitem__(self, name, value):
        if self._on_change is None or self._on_change([self.data], [name], [value]):
            self.data.__setitem__(name, value)

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __delitem__(self, k):
        raise NotImplementedError

    def __contains__(self, k):
        return k in self.data

    @property
    def __dict__(self):
        return self.to_dict()

    def __repr__(self):
        return repr(self.data)

    def to_dict(self):
        return self.data


def resource_path(file_path):
    """Get the absolute path to the resource, works for development and PyInstaller build."""
    return os.path.dirname(file_path)


def fixeddict_serializer(obj):
    if isinstance(obj, PropertyDict):
        return obj.to_dict()  # Return the underlying dictionary
    return obj


class FixedDict(PropertyDict):

    def __init__(self, data):
        super().__init__(None, data)

    def __setitem__(self, k, v):
        if k not in self.data:
            raise KeyError(k)

        super().__setitem__(k, v)

    def __delitem__(self, k):
        raise NotImplementedError

    def __getitem__(self, k):
        return super().__getitem__(k)


class UnitDict(FixedDict):
    def __init__(self, unit: str, value: Optional[float]):
        super().__init__({'unit': unit, 'value': value})

    def set_value_unit(self, value: float):
        self.__setitem__('value', value)

    def set_value(self, value: float):
        self.__setitem__('value', value)

    def set_unit(self, unit: str):
        self.__setitem__('unit', unit)


def add_to_dict(obj: dict, key: str, val: any) -> str:
    origen_key = key
    idx = 1
    while key in obj and obj[key] is not None:
        key = "{}.{}".format(origen_key, idx)
        idx += 1

    obj[key] = val
    return key


def parse_generic_field(field: dict) -> dict[str, any]:
    sub_fields = field.get('sub_fields')
    if type(sub_fields) is list and len(sub_fields) > 0:
        field_mapping = {'__field': field.get('field')}
        field_vals = None
        if field.get('type') == 'input-group':
            field_vals = []
            for sub_field in sub_fields:
                temp_sub_field = parse_generic_field(sub_field)
                field_vals.append(temp_sub_field.get('values'))

        elif field.get('type') == 'table':
            value_list = field.get('sub_values', [])
            field_vals = [{} for x in value_list]
            for sub_field in sub_fields:
                add_to_dict(field_mapping, sub_field.get('col_name'), sub_field.get('id'))
                for (idx, value) in enumerate(value_list):
                    add_to_dict(field_vals[idx], sub_field.get('col_name'), value[sub_field.get('id')])

        return {'values': field_vals, 'obj_mapping': field_mapping}
    elif field.get('type') == 'system-defined':
        return {'values': {'value': field.get('value'), 'unit': field.get('value_system')},
                'obj_mapping': field.get('id', field.get('field'))}
    elif field.get('type') == 'drag_element':
        return {'values': {'value': field.get('value')}, 'obj_mapping': field.get('id', field.get('field'))}
    return {'values': field.get('value'), 'obj_mapping': field.get('id', field.get('field'))}


def parse_generic_layer(key: str, layer: dict) -> dict[str, dict]:
    temp_layer = {}
    temp_id_layer = {'__key': key}
    fields = layer.get('fields', [])
    fields.sort(key=lambda x: x.get('position'))
    for field in fields:
        if field.get('type') != 'wf-next':
            temp_field = parse_generic_field(field)
            field_name = field.get('label') if len(field.get('label')) > 0 else field.get('field')
            key = add_to_dict(temp_layer, field_name, temp_field.get('values'))
            temp_id_layer[key] = temp_field.get('obj_mapping')

    return {'values': temp_layer, 'obj_mapping': temp_id_layer}


def parse_generic_object_json(segment_json_data: dict) -> dict:
    temp_segment = {}
    temp_id_segment = {'__id': segment_json_data.get('id')}
    layers = OrderedDict(
        sorted(segment_json_data.get('properties', {}).get('layers', {}).items(), key=lambda x: x[1].get('position')))
    for key, layer in layers.items():
        parse_generic_object_layer_json(key, layer, temp_segment, temp_id_segment)
    return {'values': temp_segment, 'obj_mapping': temp_id_segment,
            'validator': GenericFildValidator(segment_json_data, temp_id_segment)}


def parse_generic_object_layer_json(key, layer, values, obj_mapping):
    temp_layer = parse_generic_layer(key, layer)
    layer_name = layer.get('label') if len(layer.get('label')) > 0 else layer.get('key')
    key = add_to_dict(values, layer_name, temp_layer.get('values'))
    obj_mapping[key] = temp_layer.get('obj_mapping')


def clean_generic_field(field_obj: dict, values: any, field_mapping: dict | str = None) -> dict[str, any]:
    sub_fields = field_obj.get('sub_fields')
    if type(sub_fields) is list and len(sub_fields) > 0:
        if field_obj.get('type') == 'input-group':
            for (idx, val) in enumerate(values):
                clean_generic_field(sub_fields[idx], val)

        elif field_obj.get('type') == 'table':
            field_obj['sub_values'] = value_list = field_obj.get('sub_values', [])
            while len(value_list) < len(values):
                value_list.append({'id': uuid.uuid4().__str__()})
            for (k, v) in enumerate(values):
                for (k1, v1) in v.items():
                    value_list[k][field_mapping[k1]] = v1

    elif field_obj.get('type') == 'system-defined':
        field_obj['value'] = values['value']
        field_obj['value_system'] = values['unit']
    elif field_obj.get('type') == 'drag_element':
        field_obj['value'] = values['value']
    else:
        field_obj['value'] = values
    return field_obj


def clean_generic_object_json(segment_json_data: dict, values: dict, mapping: dict):
    for (k, v) in values.items():
        layer_mapping = mapping.get(k, {})
        layer_key = layer_mapping.get('__key')
        for (value_name, value_ob) in v.items():
            field_mapping = layer_mapping.get(value_name)
            field_key = ''
            if type(field_mapping) is str:
                field_key = field_mapping
            elif type(field_mapping) is dict:
                field_key = field_mapping.get('__field')
            fields = segment_json_data.get('properties').get('layers').get(layer_key).get('fields')
            field_obj = next((x for x in fields if x.get('field') == field_key), None)
            clean_generic_field(field_obj, value_ob, field_mapping)


def merge_dicts(*args) -> dict:
    dicts = list(args)
    if len(dicts) == 0:
        return {}
    res = dicts.pop(0)
    for d in dicts:
        _merge_dicts(res, d)
    return res


def _merge_dicts(a: dict | list, b: dict | list, path: list = None):
    if path is None:
        path = []

    if isinstance(b, dict):
        iterator = b.items()
    elif isinstance(b, list):
        iterator = enumerate(b)
    else:
        return

    def get_values(elem: list | dict, idx: str | int) -> any:
        if isinstance(elem, dict):
            return elem.get(idx)
        elif isinstance(elem, list):
            try:
                return elem[idx]
            except:
                pass
        return None

    def set_values(elem: list | dict, idx: str | int, value: any):
        if isinstance(elem, dict):
            elem[idx] = value
            return
        elif isinstance(elem, list):
            try:
                elem[idx] = value
            except:
                elem.append(value)

    for key, b_val in iterator:
        a_val = get_values(a, key)
        if isinstance(a_val, dict) and isinstance(b_val, dict):
            _merge_dicts(a_val, b_val, path + [str(key)])
        elif isinstance(a_val, list) and isinstance(b_val, list):
            _merge_dicts(a_val, b_val, path + [str(key)])
        else:
            set_values(a, key, b_val)
    return a


def snake_to_camel_case(snake_str):
    return "".join(x.capitalize() for x in snake_str.lower().split("_"))


class TypedList(list):
    def __init__(self, element_type, *args):
        self.element_type = element_type
        super().__init__(*args)
        self._check_elements()

    def _prepare_element(self, element):
        return element

    def after_append(self, element):
        pass

    def _check_elements(self):
        for element in self:
            self._check_element(element)

    def _check_element(self, element):
        if not isinstance(element, self.element_type):
            raise TypeError(f"All elements must be of type {self.element_type.__name__}")

    def append(self, element):
        self._check_element(element)
        prepared_element = self._prepare_element(element)
        super().append(prepared_element)
        self.after_append(prepared_element)
        return prepared_element

    def extend(self, iterable):
        for element in iterable:
            self.append(element)

    def insert(self, index, element):
        self._check_element(element)
        prepared_element = self._prepare_element(element)
        super().insert(index, prepared_element)
        self.after_append(prepared_element)

    def __setitem__(self, index, element):
        if isinstance(index, slice):
            for i in element:
                self._check_element(element)
        else:
            self._check_element(element)
        super().__setitem__(index, self._prepare_element(element))

    def __add__(self, other):
        x = copy.copy(self)
        x.extend(other)
        return x

    def __iadd__(self, other):
        self.extend(other)
        return self


def quill_hedging(input: str | dict, name: str) -> dict[str:list]:
    if isinstance(input, str):
        return {
            'ops': [
                {'insert': input}
            ]
        }
    if not isinstance(input, (dict, PropertyDict)) or 'ops' not in input or not isinstance(input['ops'], list):
        raise ValueError(f'{name} must be in the Quill.js format. (https://quilljs.com/docs/delta)')
    return input



import json

import requests
from dateutil.parser import parse as parsedate
from requests.exceptions import ConnectionError as CError
from chemotion_api.schema_manager import SchemaManager


class Connection(requests.Session):
    def __init__(self, host_url: str, verify_ssl: bool = True):
        super().__init__()
        self._host_url = host_url.strip('/')
        self._verify = verify_ssl
        self._bearer_token = None
        self.tzinfo = None
        self._schema_manager = None

    @property
    def schema_manager(self) -> SchemaManager:
        if self._schema_manager is None:
            res = self.get('/about')
            self._schema_manager = SchemaManager(res)
        return self._schema_manager


    def reset_bearer_token(self):
        self._bearer_token = None

    def set_bearer_token(self, token: str):
        self._bearer_token = token

    @property
    def bearer_token(self):
        return self._bearer_token

    @property
    def host_url(self):
        return self._host_url

    def delete(self, url_path: str = '', **kwargs) -> requests.Response:
        return self._send_request('delete', url_path, self.get_default_session_header(), kwargs)

    def get(self, url_path: str = '', **kwargs) -> requests.Response:
        return self._send_request('get', url_path, self.get_default_session_header(), kwargs)

    def post(self, url_path: str = '', **kwargs) -> requests.Response:
        return self._send_request('post', url_path, self.get_json_session_header(), kwargs)

    def patch(self, url_path: str = '', **kwargs) -> requests.Response:
        return self._send_request('patch', url_path, self.get_json_session_header(), kwargs)

    def put(self, url_path: str = '', **kwargs) -> requests.Response:
        return self._send_request('put', url_path, self.get_json_session_header(), kwargs)

    def _send_request(self, method: str, url_path: str, default_header: dict, kwargs: dict) -> requests.Response:
        kwargs['verify'] = kwargs.get('verify', self._verify)
        kwargs['url'] = kwargs.get('url', f"{self._host_url}/{url_path.lstrip('/')}")
        kwargs['headers'] = kwargs.get('headers', default_header)
        kwargs['timeout'] = kwargs.get('timeout', (2, 500))
        if 'files' in kwargs:
            if 'Content-Type' in kwargs['headers']:
                del kwargs['headers']['Content-Type']
        elif 'data' in kwargs and not isinstance(kwargs['data'], str) and kwargs['headers'].get(
                'Content-Type') == 'application/json':
            kwargs['data'] = json.dumps(kwargs.get('data', {}))
        try:
            res = getattr(super(), method)(**kwargs)
        except CError as e:
            raise ConnectionError(str(e))
        if self.tzinfo is None:
            if 'Date' in res.headers:
                self.tzinfo = parsedate(res.headers['Date']).astimezone().tzinfo

        return res

    def get_default_session_header(self) -> dict[str, str]:
        headers = {'User-Agent': 'Mozilla/5.0'}
        if self._bearer_token is not None:
            headers["Authorization"] = f"Bearer {self._bearer_token}"
        return headers

    def get_json_session_header(self) -> dict[str, str]:
        header = self.get_default_session_header()
        header['Content-Type'] = 'application/json'
        return header



from abc import ABC, abstractmethod
from typing import Type


def all_instances_of(lst: list, cls: Type):
    return isinstance(lst, list) and all(isinstance(item, cls) for item in lst)

class ChemotionSerializable(ABC):
    @abstractmethod
    def to_dict(self) -> dict:
        pass


def generate_pkg_info():
    return {
        'pkg': {
            "name": "chem-generic-ui",
            "version": "1.4.4"
        }
    }


import hashlib
import json
import os
import uuid

MAX_UPLOAD_SIZE = 5000


class AnalysesManager():
    @staticmethod
    def build_new_analyses():
        json_path = os.path.join(os.path.dirname(__file__), 'empty_elements/analyses.json')
        data = {}
        if os.path.exists(json_path):
            with open(json_path, 'r') as f:
                data = json.loads(f.read())
        data['id'] = uuid.uuid4().__str__()
        return data

    @staticmethod
    def build_new_dataset():
        json_path = os.path.join(os.path.dirname(__file__), 'empty_elements/dataset.json')
        data = {}
        if os.path.exists(json_path):
            with open(json_path, 'r') as f:
                data = json.loads(f.read())
        data['id'] = uuid.uuid4().__str__()
        return data

    @staticmethod
    def upload_file(session, file_path: str, file_uuid: str):
        with open(file_path, 'rb') as f:
            body = f.read()
            file_name = os.path.basename(file_path)
            key = file_uuid
            snippet = 0
            counter = 0
            hash_md5 = hashlib.md5()
            while snippet < len(body):
                start_snippet = snippet
                snippet += MAX_UPLOAD_SIZE
                file_chunk = body[start_snippet:snippet]
                hash_md5.update(file_chunk)
                payload = {'file': (file_name, file_chunk)}
                res = session.post('/api/v1/attachments/upload_chunk', data={'key': key, 'counter': counter},
                                   files=payload)
                counter += 1
                if res.status_code == 401:
                    raise PermissionError('Not allowed to delete device (Only for super devices or admins)')
                elif (res.status_code != 200 and res.status_code != 201):
                    raise ConnectionError()
            data = {'key': key, 'filename': file_name, 'checksum': hash_md5.hexdigest()}
            res = session.post('/api/v1/attachments/upload_chunk_complete',
                               data=data)
            if res.status_code == 401:
                raise PermissionError('Not allowed to delete device (Only for super devices or admins)')
            elif (res.status_code != 200 and res.status_code != 201):
                raise ConnectionError(f'{res.status_code} -> {res.text}')

            return data



from chemotion_api.labImotion.items.validation.registry import SchemaRegistry

generic_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "$id": "chemotion://generic/select_option/draft-01",
    "title": "Schema for generic selection options",
    "type": "object",
    "patternProperties": {
        "^[a-z][a-zA-Z_]{1,10}[a-z]$": {
            "type": "object",
            "properties": {
                "options": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "key": {
                                "type": "string"
                            },
                            "label": {
                                "type": "string"
                            }
                        },
                        "additionalProperties": False,
                        "required": [
                            "key",
                            "label"
                        ]
                    }
                }
            },
            "additionalProperties": False,
            "required": [
                "options"
            ]
        }
    },
    "additionalProperties": False,
}

SchemaRegistry.instance().register(generic_schema)



from chemotion_api.elements.abstract_element import AbstractElement

from chemotion_api.elements.sample import Sample
from collections.abc import MutableMapping
import string

class WellplateCol(MutableMapping):

    def __init__(self, *args, **kwargs):
        self._keys = [letter for letter in string.ascii_uppercase[0:8]]
        self.store = dict.fromkeys(self._keys, None)
        self.update(dict(*args, **kwargs))  # use the free update to set keys

    def __getitem__(self, key: str):
        return self.store[self._keytransform(key)]

    def __setitem__(self, key: str, value: Sample| None):
        self.store[self._keytransform(key)] = value.split() if value is not None else None

    def __delitem__(self, key):
        del self.store[self._keytransform(key)]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def _keytransform(self, key):
        key = key.upper()
        if key not in self._keys:
            raise IndexError("A Wallplate column only allows keys from A-H")
        return key

class Wellplate(AbstractElement):
    """
    Wellplates represent the localization of materials (chemicals)
    for biological experiments and are therefore some kind of link between chemical and biological operations.
    It extends the :class:`chemotion_api.elements.abstract_element.AbstractElement`

    - Plate size: 12 x 8
    - Verical indexing: 0-11
    - Horiconcal indexing: 'A' - 'H'

    Usage::

    >>> from chemotion_api import Instance
    >>> from chemotion_api.collection import Collection
    >>> import logging
    >>> try:
    >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
    >>> except ConnectionError as e:
    >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
    >>> # Get the reaction with ID 1
    >>> wp = instance.get_wellplate(1)
    >>> # Add samples to wellplate
    >>> wp.wells[0]['A'] = instance.get_sample(1)
    >>> wp.wells[0]['B'] = instance.get_sample(1)
    >>> wp.wells[0]['C'] = instance.get_sample(1)
    >>> wp.wells[1]['A'] = instance.get_sample(2)
    >>> wp.wells[1]['B'] = instance.get_sample(2)
    >>> wp.wells[1]['C'] = instance.get_sample(2)
    >>> wp.save()

    """

    def _set_json_data(self, json_data):
        super()._set_json_data(json_data)
        self.wells = [WellplateCol() for i in range(12)]
        for element in self.json_data.get('wells'):
            if element.get('sample') is not None:
                x = int(element['position']['x']) - 1
                y = chr(int(element['position']['y']) + ord('A') - 1)
                self.wells[x][y] = Sample(self._generic_segments, self._session, element['sample'])

    def _parse_properties(self) -> dict:

        return {
            'name': self.json_data.get('name'),
            'description': self.json_data.get('description')
        }

    def _clean_properties_data(self, serialize_data : dict | None =None) -> dict:
        self.json_data['wells'] = self.json_data.get('wells')[:self.json_data.get('size')]
        for element in self.json_data.get('wells'):
            x = int(element['position']['x']) - 1
            y = chr(int(element['position']['y']) + ord('A') - 1)
            element['is_new'] = element.get('is_new', False)
            if self.wells[x][y] is None:
                element['sample'] = None
            else:
                element['sample'] = self.wells[x][y].clean_data()


        self.json_data['name'] = self.properties.get('name')
        self.json_data['description'] = self.properties.get('description')
        return self.json_data

    def save(self):
        super().save()
        self.load()



import re

import requests


class SchemaManager:

    iri_prefix = 'chemotion'
    def __init__(self, res: requests.Response):
        a = re.findall(r'Version: .*(\d+\.\d+.\d+)', res.text)
        if len(a) > 0:
            self.version = a[0]
        else:
            self.version = 'TEST'

    def generate_model_type(self, tpye_name: str, schema_version):
        if schema_version is not None:
            return f'{self.iri_prefix}:generic_type/{tpye_name.lower()}/{schema_version}'
        short_version = '.'.join(self.version.split('.')[:2])
        return f'{self.iri_prefix}:type/{tpye_name.lower()}/{short_version}'


ALL_SCHEMAS = ["schema_generc_segments", "schema_generc_element", "schema_generc_layer", "schema_generc_linked_element",
               "schema_select_options", "schema_properties", "schema_cond_field", "schema_flow_obj", "schema_field",
               "schema_generc_dataset"]
